// CGraphics.cpp
// by Comanche

#include "CGraphics.h"
#include "CMain.h"
#include "CLog.h"
#include "CTime.h"
#include "CLandScape.h"
#include "CVoxel.h"
#include "CCheckBox.h"
#include "CLiquid.h"

extern CMain GameMain;
extern CLog Log;
extern CTime Time;
extern CLandScape Land;
extern CVoxel Voxel;
extern CCheckBox *CheckBoxes[20];
extern CLiquid *Liquids[128];

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::Lock()
{
	GameMain.lpddsos1->Lock(NULL,&GameMain.ddsd,DDLOCK_WAIT,NULL);
	video_buffer = (BYTE*)GameMain.ddsd.lpSurface;
	WordsPerLine = (GameMain.ddsd.lPitch);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::UnLock()
{
	GameMain.lpddsos1->Unlock(NULL);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::LockBackBuffer()
{
	GameMain.lpddsback->Lock(NULL,&GameMain.ddsd,DDLOCK_WAIT,NULL);
	video_buffer = (BYTE*)GameMain.ddsd.lpSurface;
	WordsPerLine = (GameMain.ddsd.lPitch);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::UnLockBackBuffer()
{
	GameMain.lpddsback->Unlock(NULL);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::Flip()
{
	GetDC();
	if(!CheckBoxes[1]->State)
		sprintf(buffer,"%d FPS",Time.GetFPS(Time.GetFrameTime()));
	else sprintf(buffer,"%d VPF",Land.vps);
	Text(50,237,buffer);	
	if(GameMain.Editor==0)
	{
		if(!CheckBoxes[2]->State)
			sprintf(buffer,"%d,%d",Land.ScreenPosX,Land.ScreenPosY);
		else 
		{
			if(GameMain.ActiveLiquid<0)
			{				
				sprintf(buffer,"%d",GameMain.ActiveLiquid);			
			}
			else
			{
				sprintf(buffer,"%d,%d",GameMain.ActiveLiquid,
						Liquids[GameMain.ActiveLiquid]->height);
			}
		}
		Text(50,217,buffer);	
	}
	if(GameMain.Editor==1)
	{
		sprintf(buffer,"%d,%d",Voxel.ViewX,Voxel.ViewY);
		Text(50,217,buffer);	
	}
	ReleaseDC();

	GameMain.lpddsprimary->Flip(GameMain.lpddsback, DDFLIP_WAIT);
}	

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::BltToBackBuffer()
{
	GameMain.lpddsback->BltFast(0,0,GameMain.lpddsos1,NULL,DDBLTFAST_WAIT|DDBLTFAST_NOCOLORKEY);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::GetDC()
{
	GameMain.lpddsback->GetDC(&xdc);
	Arial=CreateFontIndirect(&logfont);
	if(DrawFont) SelectObject(xdc,Arial);
	SetTextColor(xdc,RGB(textcolor.rgbtGreen,textcolor.rgbtGreen,textcolor.rgbtBlue));
	SetBkMode(xdc, TRANSPARENT);
	SetTextAlign(xdc,TA_CENTER);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::GetLaDC()
{
	GameMain.lpddsback->GetDC(&xdc);
	Arial=CreateFontIndirect(&logfont);	
	if(DrawFont) SelectObject(xdc,Arial);
	SetTextColor(xdc,RGB(textcolor.rgbtGreen,textcolor.rgbtGreen,textcolor.rgbtBlue));
	SetBkMode(xdc, TRANSPARENT);
	SetTextAlign(xdc,TA_LEFT);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::ReleaseDC()
{
	GameMain.lpddsback->ReleaseDC(xdc);
	DeleteObject(Arial);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::OSGetDC()
{
	GameMain.lpddsos1->GetDC(&xdc);
	Arial=CreateFontIndirect(&logfont);
	if(DrawFont) SelectObject(xdc,Arial);
	SetTextColor(xdc,RGB(textcolor.rgbtGreen,textcolor.rgbtGreen,textcolor.rgbtBlue));
	SetBkMode(xdc, TRANSPARENT);
	SetTextAlign(xdc,TA_CENTER);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::OSGetLaDC()
{
	GameMain.lpddsos1->GetDC(&xdc);
	Arial=CreateFontIndirect(&logfont);	
	if(DrawFont) SelectObject(xdc,Arial);
	SetTextColor(xdc,RGB(textcolor.rgbtGreen,textcolor.rgbtGreen,textcolor.rgbtBlue));
	SetBkMode(xdc, TRANSPARENT);
	SetTextAlign(xdc,TA_LEFT);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::OSReleaseDC()
{
	GameMain.lpddsos1->ReleaseDC(xdc);
	DeleteObject(Arial);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::RestorePrimarySurface()
{
	if (GameMain.lpddsos1->IsLost()!=DD_OK)
		GameMain.lpddsos1->Restore();
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::RestoreOffScreenSurface()
{
	if (GameMain.lpddsprimary->IsLost()!=DD_OK)
		GameMain.lpddsprimary->Restore();
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::Blt(int x,int y,int x2,int y2,int vx, int vy)
{
	RECT source;

	source.left=x;
	source.right=x2;
	source.top=y;
	source.bottom=y2;

	GameMain.lpddsos1->BltFast(x+vx,y+vy,GameMain.lpddsos1,&source,NULL);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::Text(int x,int y,char *string)
{
	int len=strlen(string);
	TextOut(xdc,x,y,string,len);
}

//////////////////////////////////////////////////////////////////////////////////

__declspec(naked) void CGraphics::PutPixel(int x,int y,BYTE red,BYTE green,BYTE blue)
{
	__asm
	{	
		push ebp
		mov ebp,esp
		push eax
		push ebx
		mov eax,y
		mov ebx,x
		mul WordsPerLine
		lea eax,[eax+ebx*4]
		add eax,video_buffer
		mov bl,blue
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,green
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,red
		mov BYTE PTR [eax],bl
		pop ebx
		pop eax	
		mov esp,ebp
		pop ebp
		ret 14h
	}
}

//////////////////////////////////////////////////////////////////////////////////

__declspec(naked) void CGraphics::PutSafePixel(int x,int y,BYTE red,BYTE green,BYTE blue)
{
	__asm
	{	
		push ebp
		mov ebp,esp
		push eax
		push ebx
		mov ebx,x
		mov eax,y
		cmp eax,0
		jl end
		cmp eax,479
		jg end
		cmp ebx,101
		jl end
		cmp ebx,559
		jg end
		mul WordsPerLine
		lea eax,[eax+ebx*4]
		add eax,video_buffer
		mov bl,blue
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,green
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,red
		mov BYTE PTR [eax],bl
end:	pop ebx
		pop eax	
		mov esp,ebp
		pop ebp
		ret 14h
	}
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::Quader(int x,int y,int x2,int y2, 
			  BYTE red,BYTE green,BYTE blue)
{
	int i;
	for (i=x;i<=x2;i++)
	{
		PutPixel(i,y,red,green,blue);
		PutPixel(i,y2,red,green,blue);
	}
	for (i=y;i<=y2;i++)
	{
		PutPixel(x,i,red,green,blue);
		PutPixel(x2,i,red,green,blue);
	}
}

//////////////////////////////////////////////////////////////////////////////////


void CGraphics::FullQuader(int x,int y,int x2,int y2, 
				  BYTE red,BYTE green,BYTE blue)
{
	__asm
	{
		push eax
		push ebx
		push ecx
		push edx
		mov eax,x
		mov ebx,x2
		cmp eax,ebx
		jg end
		mov eax,y
		mov ebx,y2
		cmp eax,ebx
		jg end
		mov eax,y
		mov ebx,x
		mul WordsPerLine
		lea eax,[eax+ebx*4]
		add eax,video_buffer
		mov edx,x2
		sub edx,x
		mov ecx,y2
		sub ecx,y
jmp1:	push ecx
		push eax
		mov ecx,edx
jmp2:	mov bl,blue		
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,green
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,red
		mov BYTE PTR [eax],bl
		add eax,2
		loop jmp2
		pop eax
		pop ecx	
		add eax,WordsPerLine
		loop jmp1
end:	pop eax
		pop ebx
		pop edx
		pop ecx
	}
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::BlackQuader(int x,int y,int x2,int y2)
{
	DDBLTFX ddbltfx;
	RECT rect;

	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = 0;

	rect.left=x;
	rect.right=x2;
	rect.top=y;
	rect.bottom=y2;

	GameMain.lpddsos1->Blt(&rect,NULL,NULL,DDBLT_COLORFILL,&ddbltfx);
}


//////////////////////////////////////////////////////////////////////////////////

__declspec(naked) void CGraphics::PutTransPixel(int x,int y,BYTE red,BYTE green,BYTE blue)
{
	__asm
	{	
		push ebp
		mov ebp,esp
		push eax
		push ebx
		push ecx
		mov ebx,x
		mov eax,y
		mul WordsPerLine
		lea eax,[eax+ebx*4]
		add eax,video_buffer
		movzx ebx,BYTE PTR [eax]
		movzx ecx,blue
		add ebx,ecx
		shr ebx,1
		mov BYTE PTR [eax],bl
		inc eax		
		movzx ebx,BYTE PTR [eax]
		movzx ecx,green
		add ebx,ecx
		shr ebx,1
		mov BYTE PTR [eax],bl
		inc eax
		movzx ebx,BYTE PTR [eax]
		movzx ecx,red
		add ebx,ecx
		shr ebx,1
		mov BYTE PTR [eax],bl
		pop ecx
		pop ebx
		pop eax
		mov esp,ebp
		pop ebp
		ret 14h
	}
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::FullTransQuader(int x,int y,int x2,int y2,
					   BYTE red,BYTE green,BYTE blue)
{
	BYTE color[8];
	__asm
	{
		push eax
		push ebx
		push ecx
		push edx
		mov eax,x
		mov ebx,x2
		cmp eax,ebx
		jg end
		mov eax,y
		mov ebx,y2
		cmp eax,ebx
		jg end
		mov al,blue
		mov bl,green
		mov cl,red
		mov BYTE PTR [color],al
		mov BYTE PTR [color+1],bl
		mov BYTE PTR [color+2],cl
		mov BYTE PTR [color+3],0
		mov BYTE PTR [color+4],al
		mov BYTE PTR [color+5],bl
		mov BYTE PTR [color+6],cl
		mov BYTE PTR [color+7],0
		pxor mm4,mm4
		mov eax,y
		mov ebx,x
		mul WordsPerLine
		lea eax,[eax+ebx*4]
		add eax,video_buffer
		mov edx,x2
		sub edx,x
		mov ecx,y2
		sub ecx,y
		movq mm5,color
jmp1:	push ecx
		push eax
		mov ecx,edx	
jmp2:	dec ecx
		jz jmp3
		movq mm0,BYTE PTR [eax]
		movq mm1,mm0
		movq mm2,mm5
		movq mm3,mm5
		punpckhbw mm0,mm4
		punpcklbw mm1,mm4
		punpckhbw mm2,mm4
		punpcklbw mm3,mm4
		paddw mm0,mm2
		paddw mm1,mm3
		psrlw mm0,1
		psrlw mm1,1
		packuswb mm1,mm0
		movq BYTE PTR [eax],mm1
		add eax,8
		loop jmp2
		pop eax
		pop ecx	
		add eax,WordsPerLine
		loop jmp1
		jmp end
jmp3:	movd mm0,BYTE PTR [eax]
		movd mm1,BYTE PTR [eax]
		movd mm2,color
		movd mm3,color	
		punpckhbw mm0,mm4
		punpcklbw mm1,mm4
		punpckhbw mm2,mm4
		punpcklbw mm3,mm4
		paddw mm0,mm2
		paddw mm1,mm3
		psrlw mm0,1
		psrlw mm1,1
		packuswb mm1,mm0
		movd BYTE PTR [eax],mm1
		pop eax
		pop ecx	
		add eax,WordsPerLine
		loop jmp1
end:	pop eax
		pop ebx
		pop edx
		pop ecx
		emms		
	}
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::FullBrightQuader(int x,int y,int x2,int y2,BYTE red,BYTE green,BYTE blue)
{
	BYTE color[8];
	__asm
	{
		push eax
		push ebx
		push ecx
		push edx
		mov eax,x
		mov ebx,x2
		cmp eax,ebx
		jg end
		mov eax,y
		mov ebx,y2
		cmp eax,ebx
		jg end
		mov al,blue
		mov bl,green
		mov cl,red
		mov BYTE PTR [color],al
		mov BYTE PTR [color+1],bl
		mov BYTE PTR [color+2],cl
		mov BYTE PTR [color+3],0
		mov BYTE PTR [color+4],al
		mov BYTE PTR [color+5],bl
		mov BYTE PTR [color+6],cl
		mov BYTE PTR [color+7],0
		mov eax,y
		mov ebx,x
		mul WordsPerLine
		lea eax,[eax+ebx*4]
		add eax,video_buffer
		mov edx,x2
		sub edx,x
		mov ecx,y2
		sub ecx,y
		movq mm2,color
jmp1:	push ecx
		push eax
		mov ecx,edx	
jmp2:	dec ecx
		jz jmp3
		movq mm0,BYTE PTR [eax]
		movq mm1,mm2
		paddusb mm0,mm1
		movq BYTE PTR [eax],mm0
		add eax,8
		loop jmp2
		pop eax
		pop ecx	
		add eax,WordsPerLine
		loop jmp1
		jmp end
jmp3:	movd mm0,BYTE PTR [eax]
		movd mm1,mm2
		paddusb mm0,mm1
		movq BYTE PTR [eax],mm0
		pop eax
		pop ecx	
		add eax,WordsPerLine
		loop jmp1
end:	pop eax
		pop ebx
		pop edx
		pop ecx
		emms		
	}
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::Line(int x,int y,int x2,int y2,BYTE red,BYTE green,BYTE blue)
{
	int lenx,leny;
	
	if(x>x2) lenx=x-x2;
	else lenx=x2-x;
	if(y>y2) leny=y-y2;
	else leny=y2-y;
	
	if (lenx>leny)
	{
		float c=(float)(lenx)/(float)(leny);
		int i=0;
		while(y<=y2)
		{
			i++;y++;
			for(int i2=0;i2<=((i+1)*c)-(i*c);i2++)
				PutPixel((x+(float)(i*c))+i2,y,red,green,blue);
		}
		return;
	}
	else
	{
		float c=(float)(leny)/(float)(lenx);
		int i=0;
		while(x<=x2)
		{
			i++;x++;
			for(int i2=0;i2<=c;i2++)
				PutPixel(x,(y+(float)(i*c))-i2,red,green,blue);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////

/*void CGraphics::Line(int x,int y,int x2,int y2,BYTE red,BYTE green,BYTE blue)
{
	if(x2>x) int len=x2-x;
	if(x2==x) int len=0;
	if(x2<x) int len=x-x2;

	if(y2>y) int len2=y2-y;
	if(y2==y) int len2=0;
	if(y2<y) int len2=y-y2;

	if(len==len2) return();

	if(len>len2)
	{
		if(y>y2) c=y2;
		if(y<y2) c=y;

		for(int i=1; i<=len; i++)
		{
			float b=len2/i;
					
			PutPixel(i,b,red,green,blue);
		}		
	}

}*/

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::HLine(int x,int y,int p,BYTE red,BYTE green,BYTE blue)
{
	__asm
	{
		push eax
		push ebx
		push ecx
		mov ebx,x
		mov eax,y
		shl ebx,2				
		mul WordsPerLine
		add eax,ebx
		add eax,video_buffer
		mov ecx,p
		inc ecx
jmp1:	mov bl,blue		
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,green
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,red
		mov BYTE PTR [eax],bl
		sub eax,2
		add eax,WordsPerLine
		loop jmp1
		pop ecx
		pop ebx
		pop eax	
	}
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::VLine(int x,int y,int p,BYTE red,BYTE green,BYTE blue)
{
	__asm
	{
		push eax
		push ebx
		push ecx
		mov ebx,x
		mov eax,y
		shl ebx,2				
		mul WordsPerLine
		add eax,ebx
		add eax,video_buffer
		mov ecx,p		
jmp1:	mov bl,blue		
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,green
		mov BYTE PTR [eax],bl
		inc eax
		mov bl,red
		mov BYTE PTR [eax],bl
		add eax,2
		loop jmp1
		pop ecx
		pop ebx
		pop eax	
	}
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::TransHLine(int x,int y,int p,BYTE red,BYTE green,BYTE blue)
{
	for(int i=0;i<=p;i++)
		PutTransPixel(x,y+i,red,green,blue);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::TransVLine(int x,int y,int p,BYTE red,BYTE green,BYTE blue)
{
	for(int i=0;i<=p;i++)
		PutTransPixel(x+i,y,red,green,blue);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::DrawButton(int x,int y,int x2,int y2)
{
	FullQuader(x,y,x2,y2,150,150,150);
	VLine(x,y,x2-x,200,200,200);
	VLine(x+1,y+1,x2-x-2,170,170,170);
	HLine(x,y,y2-y-1,200,200,200);
	HLine(x+1,y+1,y2-y-3,170,170,170);
	VLine(x,y2,x2-x,10,10,10);
	VLine(x+1,y2-1,x2-x-2,80,80,80);
	HLine(x2,y,y2-y,10,10,10);
	HLine(x2-1,y+1,y2-y-2,80,80,80);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::DrawPushedButton(int x,int y,int x2,int y2)
{
	FullQuader(x+1,y+1,x2-1,y2-1,150,150,150);
	VLine(x,y,x2-x,10,10,10);
	VLine(x+1,y+1,x2-x-1,80,80,80);
	HLine(x,y,y2-y,10,10,10);
	HLine(x+1,y+1,y2-y-2,80,80,80);
	VLine(x,y2,x2-x,200,200,200);
	VLine(x+1,y2-1,x2-x-2,170,170,170);
	HLine(x2,y,y2-y,200,200,200);
	HLine(x2-1,y+1,y2-y-2,190,170,170);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::DrawCheckBoxOff(int x,int y)
{
	FullQuader(x+1,y+1,x+10,y+10,220,220,220);
	VLine(x,y,11,10,10,10);
	VLine(x+1,y+1,10,80,80,80);
	HLine(x,y,11,10,10,10);
	HLine(x+1,y+1,9,80,80,80);
	VLine(x,y+11,11,200,200,200);
	VLine(x+1,y+10,9,170,170,170);
	HLine(x+11,y,11,200,200,200);
	HLine(x+10,y+1,9,190,170,170);
}

//////////////////////////////////////////////////////////////////////////////////

void CGraphics::DrawCheckBoxOn(int x,int y)
{
	FullQuader(x+1,y+1,x+10,y+10,220,220,220);
	VLine(x,y,11,10,10,10);
	VLine(x+1,y+1,10,80,80,80);
	HLine(x,y,11,10,10,10);
	HLine(x+1,y+1,9,80,80,80);
	VLine(x,y+11,11,200,200,200);
	VLine(x+1,y+10,9,170,170,170);
	HLine(x+11,y,11,200,200,200);
	HLine(x+10,y+1,9,190,170,170);
	for(int i=0;i<5;i++)
	{
		PutPixel(x+3+i,y+4+i,75,75,75);
		PutPixel(x+4+i,y+3+i,75,75,75);
		PutPixel(x+8-i,y+4+i,75,75,75);
		PutPixel(x+7-i,y+3+i,75,75,75);
	}
	PutPixel(x+3,y+3,75,75,75);
	PutPixel(x+3,y+8,75,75,75);
	PutPixel(x+8,y+3,75,75,75);
	PutPixel(x+8,y+8,75,75,75);
	for(i=0;i<4;i++)
	{
		PutPixel(x+4+i,y+4+i,0,0,0);
		PutPixel(x+7-i,y+4+i,0,0,0);
	}
}

//////////////////////////////////////////////////////////////////////////////////

BOOL CGraphics::StoreScreenArea(int x,int y,int x2,int y2)
{
	x2++;y2++;
	
	Sx=x;Sy=y;Sx2=x2;Sy2=y2;

	memset(&GameMain.ddsd,0,sizeof(GameMain.ddsd));
	GameMain.ddsd.dwSize = sizeof(GameMain.ddsd);
	GameMain.ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH; 
	GameMain.ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	GameMain.ddsd.dwWidth = x2-x;
	GameMain.ddsd.dwHeight = y2-y;

	if(GameMain.lpdd2->CreateSurface(&GameMain.ddsd,&lpddsos2,NULL)!=DD_OK)
	{
		Log.WriteErrorToLogFile("> Error: Store Screen Area");
		return(FALSE);
	}

	RECT source;

	source.left=x;
	source.right=x2;
	source.top=y;
	source.bottom=y2;

	if(lpddsos2->BltFast(0,0,GameMain.lpddsos1,&source,NULL)!=DD_OK)
	{
		lpddsos2->Release();
		Log.WriteErrorToLogFile("> Error: Store Screen Area");
		return(FALSE);
	}

	return(TRUE);
}

//////////////////////////////////////////////////////////////////////////////////

BOOL CGraphics::RestoreScreenArea()
{
	if(GameMain.lpddsos1->BltFast(Sx,Sy,lpddsos2,NULL,NULL)!=DD_OK)
	{
		Log.WriteErrorToLogFile("> Error: Restore Screen Area");
		lpddsos2->Release();
		return(FALSE);
	}

	if(lpddsos2!=NULL)
	{
		lpddsos2->Release();
		lpddsos2=NULL;
	}

	return(TRUE);
}